﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ConsoleApplication5
{
    sealed class Manager
    {
        IO io = new IO();


        private Town[] towns;                   // Jednotlivá města
        private TypeOfRoad[] typesOfRoad;       // Typy silnic
        private Condition[] conditions;         // Podmínky
        private Landscape[] landscapes;         // Krajiny
        private int MaxFinance;                 // Maximální finance

        private List<Road> T = new List<Road>();                    // Množina T se všemi výslednými silnicemi
        private List<Road> roads = new List<Road>();                // List se všemi možnostmi silnic
        private List<Component> components = new List<Component>(); // Komponenty propojených měst silnicemi

        // Počítek vytváření silnic
        public void StartAlogorithm()
        {
            SortTypesOfRoad();              // Seřadí typy silnic v poli od nejpomalejší silnice po nejrychlejší

            CreateRoads();                  // Vypočítání a následné přiřazení silnic z daného města k jiným městům

            AddTownsToComponents();         // Přidá města do komponent

            // CreateConditionRoads();         // Nejprve se vytvoří silnice v podmínce

            // (BORŮVKŮV ALGORITMUS)
            // Dokud je komponent víc než 1
            while (components.Count > 1)
            {
                List<Road> E = new List<Road>();

                List<Component> remove = new List<Component>();

                foreach (Component component in components)     // Pro každou komponentu
                {
                    if (!remove.Contains(component))
                    {
                        List<Road> S = new List<Road>();
                        Road fastest = null;
                        // Pro každé město z dané komponenty
                        foreach (Town town in component.Towns)
                        {
                            fastest = town.Roads[0];
                            // Určí nejrychlejší trasu z jednoho města do druhého
                            foreach (Road road in town.Roads)
                            {
                                if (road.Time < fastest.Time)
                                {
                                    fastest = road;
                                }
                            }
                            S.Add(fastest);
                        }
                        fastest = S[0];
                        // Určí nejrychlejší trasu z jedné komponenty do druhé
                        foreach (Road road in S)
                        {
                            if (road.Time < fastest.Time)
                            {
                                fastest = road;
                            }
                        }
                        E.Add(fastest);

                        // Udělá ze dvou komponent jednu
                        for (int j = 0; j < components.Count; j++)
                        {
                            bool a = false;
                            foreach (Town town in components[j].Towns)
                            {
                                if (Object.Equals(town, fastest.Town2))
                                {
                                    component.Towns.AddRange(components[j].Towns);
                                    remove.Add(components[j]);
                                    a = true;
                                    break;
                                }
                            }
                            if (a) break;
                        }
                    }
                }
                T.AddRange(E);

                // Smaže komponenty
                foreach (Component c in remove)
                {
                    components.Remove(c);
                }
            }
            ModifyRoads();
        }

        private bool Contains(List<Town> list, Road road)
        {
            foreach (Town town in list)
            {
                if (Object.Equals(town, road.Town1) && Object.Equals(town, road.Town2))
                {
                    return true;
                }
            }
            return false;
        }

        private bool Contains(List<Component> list, Component component)
        {
            foreach (Component comp in list)
            {
                if (Object.Equals(comp, component))
                {
                    return true;
                }
            }
            return false;
        }

        // Seřadí typy silnic v poli od nejlevnější silnice po nejdražší silnici
        private void SortTypesOfRoad()
        {
            List<TypeOfRoad> typesOfRoad = new List<TypeOfRoad>(this.typesOfRoad);
            TypeOfRoad[] typesOfRoad2 = new TypeOfRoad[this.typesOfRoad.Count()];

            for (int i = 0; i < this.typesOfRoad.Count(); i++)
            {
                TypeOfRoad cheapest = typesOfRoad[0];

                foreach (TypeOfRoad typeOfRoad in typesOfRoad)
                {
                    if (typeOfRoad.Price < cheapest.Price)
                    {
                        cheapest = typeOfRoad;
                    }
                }

                typesOfRoad2[i] = cheapest;
                typesOfRoad.Remove(cheapest);

            }
            this.typesOfRoad = typesOfRoad2;
        }

        // Vytvoří a přidá ke každému městu všechny možné silnice do jiných měst
        // Ošetřeno tak, aby se nevytvořila dvakrát stejná silnice
        private void CreateRoads()
        {
            for (int i = 0; i < towns.Count() - 1; i++)
            {
                for (int j = i + 1; j < towns.Count(); j++)
                {
                    Road road = new Road(towns[i], towns[j], typesOfRoad[0], landscapes);

                    roads.Add(road);

                    towns[i].AddRoad(road);
                    towns[j].AddRoad(road);
                }
            }
        }

        // Přidá města do komponent
        private void AddTownsToComponents()
        {
            foreach (Town town in towns)
            {
                Component component = new Component();
                component.AddTown(town);

                components.Add(component);
            }
        }

        private void ModifyRoads()
        {
            while (CheckConditions()) // Pokud je hodnota true, tak není splněna nějaká podmínka nebo finance
            {
                foreach (Condition condition in conditions)
                {
                    Town town1 = towns[condition.Index_Town1];
                    Town town2 = towns[condition.Index_Town2];

                    List<Road> path = GetPath(null, town1, town2);    // Seznam silnic, které vedou z jednoho města do druhého
                    condition.AddRoads(path);

                    float TotalTime = this.TotalTime(path);           // Celkový čas cesty mezi podmínkovými městy
                    float TotalFinance = this.TotalFinance(T);        // Celkové finance celé silniční sítě

                    if (TotalTime > condition.MaxTime) // Pokud nejsou splněny podmínky
                    {
                        if (path.Count > 1)
                        {
                            // Úprava vzdálenosti - tvorba nových silnic, aby cesta z města A do města B byla kratší
                            int ii = 0;
                            while (path.Count > 1)
                            {
                                int i = 1 + ii;
                                
                                Road road1 = new Road(path[i].Town1, towns[condition.Index_Town2], typesOfRoad[0], landscapes);
                                Road road2 = new Road(path[i].Town2, towns[condition.Index_Town2], typesOfRoad[0], landscapes);
                                Road road = null;
                                
                                if (!Contains(T, road1) && road1.Time < road2.Time + path[i].Time)
                                {
                                    road = road1;
                                }
                                if (!Contains(T, road2) && road1.Time > road2.Time + path[i].Time)
                                {
                                    road = road2;
                                }
                                
                                if (road == null)
                                {
                                    ii++;
                                    continue;
                                }

                                if (!road.IntersectRoads(T))
                                {
                                    condition.AddRoad(road);
                                    T.Add(road);
                                }
                                else
                                {
                                    bool a = false;
                                    int index = T.Count;
                                    for (int j = 0; j < index; j++)
                                    {
                                        if (!Object.Equals(road, T[j]) &&
                                            Manager.LineIntersectsLine(road.Town1.ToPoint(), road.Town2.ToPoint(),
                                                                       T[j].Town1.ToPoint(), T[j].Town2.ToPoint()) &&
                                            !Object.Equals(road1.Town1, T[j].Town1) && !Object.Equals(road1.Town1, T[j].Town2))
                                        {
                                            Road r1 = new Road(road.Town1, T[j].Town1, typesOfRoad[0], landscapes);
                                            Road r2 = new Road(road.Town1, T[j].Town2, typesOfRoad[0], landscapes);
                                            Road r = null;

                                            if (!Contains(T, r1) && r1.Time < r2.Time + T[j].Time)
                                            {
                                                r = r1;
                                            }
                                            if (!Contains(T, r2) && r1.Time > r2.Time + T[j].Time)
                                            {
                                                r = r2;
                                            }

                                            if (r == null) continue;
                                            
                                            condition.AddRoad(road);
                                            T.Add(r);
                                            a = true;
                                            RemoveExcessRoads();
                                            break;
                                        }
                                    }
                                    if (a) break; 
                                } 
                                path = GetPath(null, town1, town2);
                                if (this.TotalTime(path) < condition.MaxTime)
                                {
                                    break;
                                }
                            }
                            // Úprava času - změna typů silnic
                            foreach (TypeOfRoad typeOfRoad in typesOfRoad)
                            {
                                bool a = false;
                                for (int i = 0; i < path.Count; i++)
                                {
                                    path[i].typeOfRoad = typeOfRoad;
                                    if (this.TotalTime(path) <= condition.MaxTime)
                                    {
                                        a = true;
                                        break;
                                    }
                                }
                                if (a) break;
                            }
                        }
                        else // Úprava času - změna typu silnice
                        {
                            foreach (TypeOfRoad typeOfRoad in typesOfRoad)
                            {
                                path[0].typeOfRoad = typeOfRoad;
                                if (path[0].Time <= condition.MaxTime)
                                {
                                    break;
                                }
                            }

                        }
                    }
                }
                if (this.TotalFinance(T) > MaxFinance) // Pokud nejsou splněny finance
                {                                      // Odstraní silnice, které nepotřebujeme
                    List<Road> r = new List<Road>(); // Silnice, které nepotřebujeme

                    for (int i = 0; i < T.Count; i++)
                    {
                        bool contains = false;

                        foreach (Condition c in conditions)
                        {
                            if (c.Contains(T[i])) { contains = true; }
                        }

                        if (!contains && RoadsCount(T[i].Town1) > 1 && RoadsCount(T[i].Town2) > 1)
                        {
                            r.Add(T[i]);
                        }
                    }

                    while (this.TotalFinance(T) > MaxFinance)
                    {
                        foreach (Road road in r)
                        {
                            if (this.TotalFinance(T) > MaxFinance)
                            {
                                T.Remove(road);
                            }
                        }
                    }
                }
            }
        }

        private bool Contains(List<Road> T, Road road1)
        {
            foreach (Road road in T)
            {
                if ((Object.Equals(road.Town1, road1.Town1) && Object.Equals(road.Town2, road1.Town2)) ||
                    (Object.Equals(road.Town1, road1.Town2) && Object.Equals(road.Town2, road1.Town1)))
                {
                    return true;
                }
            }
            return false;
        }

        private void RemoveExcessRoads()
        {
            List<Road> r = new List<Road>(); // Silnice, které nepotřebujeme

            for (int i = 0; i < T.Count; i++)
            {
                bool contains = false;

                foreach (Condition c in conditions)
                {
                    if (c.Contains(T[i])) { contains = true; }
                }

                if (!contains && RoadsCount(T[i].Town1) > 1 && RoadsCount(T[i].Town2) > 1)
                {
                    r.Add(T[i]);
                }
            }

            foreach (Road road in r)
            {
                T.Remove(road);
            }
        }

        private int RoadsCount(Town town)
        {
            int RoadsCount = 0;

            foreach (Road road in T)
            {
                if (Object.Equals(town, road.Town1) || Object.Equals(town, road.Town2))
                {
                    RoadsCount++;
                }
            }

            return RoadsCount;
        }

        private bool CheckConditions()
        {
            if (TotalFinance(T) > MaxFinance) return true;

            foreach (Condition condition in conditions)
            {
                Town town1 = towns[condition.Index_Town1];
                Town town2 = towns[condition.Index_Town2];

                List<Road> path = GetPath(null, town1, town2);

                if (TotalTime(path) > condition.MaxTime)
                {
                    return true;
                }
            }
            return false;
        }

        // Vrátí celkové finance
        private float TotalFinance(List<Road> roads)
        {
            float TotalFinance = 0;

            foreach (Road road in roads)
            {
                TotalFinance += road.Finance;
            }

            return TotalFinance;
        }

        // Vrátí celkový čas cesty z jednoho města do druhého
        private float TotalTime(List<Road> roads)
        {
            float TotalTime = 0;

            foreach (Road road in roads)
            {
                TotalTime += road.Time;
            }

            return TotalTime;
        }

        // Vrátí list s městy, kterými prochází nejrychlejší cesta z jednoho města do druhého
        private List<Road> GetPath(Town last, Town from, Town to)
        {
            List<List<Road>> lists = new List<List<Road>>();

            foreach (Road road in T)
            {
                if (Object.Equals(road.Town1, from) && !Object.Equals(road.Town2, last))
                {
                    if (Object.Equals(road.Town2, to))
                    {
                        List<Road> l = new List<Road>();
                        l.Add(road);
                        return l;
                    }
                    else
                    {
                        List<Road> l = new List<Road>();
                        l.AddRange(GetPath(road.Town1, road.Town2, to));
                        if (l.Count != 0)
                        {
                            l.Add(road);
                            lists.Add(l);
                        }
                    }
                }
                else if (Object.Equals(road.Town2, from) && !Object.Equals(road.Town1, last))
                {
                    if (Object.Equals(road.Town1, to))
                    {
                        List<Road> l = new List<Road>();
                        l.Add(road);
                        return l;
                    }
                    else
                    {
                        List<Road> l = new List<Road>();
                        l.AddRange(GetPath(road.Town2, road.Town1, to));
                        if (l.Count != 0)
                        {
                            l.Add(road);
                            lists.Add(l);
                        }
                    }
                }
            }
            return ShortestPath(lists);
        }

        private List<Road> ShortestPath(List<List<Road>> Lists)
        {
            if (Lists.Count != 0)
            {
                List<Road> Shortest = Lists[0];

                foreach (List<Road> list in Lists)
                {
                    if (TotalTime(list) < TotalTime(Shortest))
                    {
                        Shortest = list;
                    }
                }

                return Shortest;
            }
            else
            {
                return new List<Road>();
            }
        }

        public static bool LineIntersectsLine(Point l1p1, Point l1p2, Point l2p1, Point l2p2)
        {
            float q = (l1p1.Y - l2p1.Y) * (l2p2.X - l2p1.X) - (l1p1.X - l2p1.X) * (l2p2.Y - l2p1.Y);
            float d = (l1p2.X - l1p1.X) * (l2p2.Y - l2p1.Y) - (l1p2.Y - l1p1.Y) * (l2p2.X - l2p1.X);

            if (d == 0)
            {
                return false;
            }

            float r = q / d;

            q = (l1p1.Y - l2p1.Y) * (l1p2.X - l1p1.X) - (l1p1.X - l2p1.X) * (l1p2.Y - l1p1.Y);
            float s = q / d;

            if (r < 0 || r > 1 || s < 0 || s > 1)
            {
                return false;
            }

            return true;
        }

        // Volá metodu k načtení input.txt ze třídy IO
        public void LoadFile()
        {
            Console.WriteLine("Zadej absolutní cestu k souboru s názvem: ");
            string absolutePath = Console.ReadLine();
            io.LoadFile(absolutePath, ref typesOfRoad, ref towns, ref landscapes, ref conditions, ref MaxFinance);
        }

        public void SaveFile()
        {
            io.SaveFile(T);
        }

    }
}
