﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ConsoleApplication5
{
    sealed class Road
    {
        public Town Town1 { get; private set; }
        public Town Town2 { get; private set; }

        private TypeOfRoad TypeOfRoad;
        public TypeOfRoad typeOfRoad {
            get
            {
                return TypeOfRoad;
            }
            set 
            {
                TypeOfRoad = value;

                Finance = DistanceTowns * TypeOfRoad.Speed;
                Time = CalcTime(TypeOfRoad, landscapes);
            }
        }
        private Landscape[] landscapes;

        public float Finance { get; private set; }
        public float DistanceTowns { get; private set; } // Vzdálenost (rozdíl) mezi městy
        public float Time { get; private set; }
        
        public Road(Town Town1, Town Town2, TypeOfRoad typeOfRoad, Landscape[] landscapes)
        {
            this.Town1 = Town1;
            this.Town2 = Town2;

            DistanceTowns = CalcDistance(Town1.ToPoint(), Town2.ToPoint());

            this.landscapes = landscapes;
            this.typeOfRoad = typeOfRoad;
        }

        private float CalcTime(TypeOfRoad typeOfRoad, Landscape[] landscapes)
        {
            float TotalTime = 0;
            float DistanceTowns = this.DistanceTowns;

            foreach (Landscape l in landscapes)
            {
                List<Point> IntersectionPoints = new List<Point>();

                Point p1 = Town1.ToPoint();
                Point p2 = Town2.ToPoint();

                if (l.Contains(p1) && l.Contains(p2))
                {
                    IntersectionPoints.Add(p1);
                    IntersectionPoints.Add(p2);
                }
                else if (l.Contains(p1))
                {
                    IntersectionPoints.Add(p1);
                }
                else if (l.Contains(p2))
                {
                    IntersectionPoints.Add(p2);
                }

                if (Manager.LineIntersectsLine(p1, p2, new Point(l.X, l.Y), new Point(l.X + l.Width, l.Y)))
                {
                    IntersectionPoints.Add(FindIntersectionPoint(p1, p2, new Point(l.X, l.Y), new Point(l.X + l.Width, l.Y)));
                }
                if (Manager.LineIntersectsLine(p1, p2, new Point(l.X + l.Width, l.Y), new Point(l.X + l.Width, l.Y + l.Height)))
                {
                    IntersectionPoints.Add(FindIntersectionPoint(p1, p2, new Point(l.X + l.Width, l.Y), new Point(l.X + l.Width, l.Y + l.Height)));
                }
                if (Manager.LineIntersectsLine(p1, p2, new Point(l.X + l.Width, l.Y + l.Height), new Point(l.X, l.Y + l.Height)))
                {
                    IntersectionPoints.Add(FindIntersectionPoint(p1, p2, new Point(l.X + l.Width, l.Y + l.Height), new Point(l.X, l.Y + l.Height)));
                }
                if (Manager.LineIntersectsLine(p1, p2, new Point(l.X, l.Y + l.Height), new Point(l.X, l.Y)))
                {
                    IntersectionPoints.Add(FindIntersectionPoint(p1, p2, new Point(l.X, l.Y + l.Height), new Point(l.X, l.Y)));
                }

                // Pokud silnice prochází krajinou
                if (IntersectionPoints.Count() == 2)
                {
                    float Distance = CalcDistance(IntersectionPoints[0], IntersectionPoints[1]); // Délka silnice procházející krajinou
                    DistanceTowns -= Distance;                                                   
                    TotalTime += (Distance * typeOfRoad.Speed * l.Time);
                }
            }
            TotalTime += (DistanceTowns * typeOfRoad.Speed);

            return TotalTime;
        }

        // Spočítá (najde) průsečík úsečky (silnice) a strany obdélníku (krajiny)
        private Point FindIntersectionPoint(Point p1, Point p2, Point p3, Point p4)
        {
            float X, Y;

            if (p3.Y == p4.Y)
            {
                Y = p3.Y;
                X = p1.X + (Y - p1.Y) * (p2.X - p1.X) / (p2.Y - p1.Y);
            }
            else
            {
                X = p3.X;
                Y = p1.Y + (X - p1.X) * (p2.Y - p1.Y) / (p2.X - p1.X);
            }

            return new Point(X, Y);
        }

        // Vypočítá vzdálenost (rozdíl) mezi dvěma body
        private float CalcDistance(Point point1, Point point2)
        {
            float x = Math.Abs(point1.X - point2.X);
            float y = Math.Abs(point1.Y - point2.Y);

            return (float)Math.Sqrt(Math.Pow(x, 2) + Math.Pow(y, 2));
        }

        public bool IntersectRoads(List<Road> roads)
        {
            foreach (Road road in roads)
            {
                if (!Object.Equals(this, road) &&
                    !Object.Equals(Town1, road.Town1) &&
                    !Object.Equals(Town1, road.Town2) &&
                    !Object.Equals(Town2, road.Town1) &&
                    !Object.Equals(Town2, road.Town2) &&
                    Manager.LineIntersectsLine(Town1.ToPoint(), Town2.ToPoint(), road.Town1.ToPoint(), road.Town2.ToPoint()))
                {
                    return true;
                }
            }
            return false;
        }

    }
}
