﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace caRoads
{
    class Condition
    {
        private Town townFrom;
        private Town townTo;
        private double timeLimit;

        public Condition(Town from, Town to, double limit)
        {
            townFrom = from;
            townTo = to;
            timeLimit = limit;
        }

        public static Condition Parse(string text, IEnumerable<Town> towns)
        {
            string[] values = text.Split(Globals.separator, StringSplitOptions.RemoveEmptyEntries);
            double limit;
            if (values.Length != 3 || !double.TryParse(values[2].Trim(), out limit))
                throw new FormatException("Vstupni parametr nema spravny format.");
            Town from = towns.FirstOrDefault(t => t.Name == values[0].Trim());
            Town to = towns.FirstOrDefault(t => t.Name == values[1].Trim());
            if (from == null || to == null)
                throw new FormatException("Vstupni parametr neobsahuje platna data.");
            return new Condition(from, to, limit);
        }

        public bool IsSatisfied(IEnumerable<Road> builtRoads)
        {
            if (!Way.ExistWay(townFrom, townTo, builtRoads))
                return false;
            Way fastestWay = Way.GetWays(townFrom, townTo, builtRoads).OrderBy(w => w.Duration).First();
            return fastestWay.Duration <= timeLimit;
        }
    }
}
