﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace caRoads
{
    public class Location
    {
        private double x;
        public double X
        {
            get { return x; }
            set
            {
                if (value < 0)
                    throw new ArgumentOutOfRangeException("Location.X");
                x = value;
            }
        }

        private double y;
        public double Y
        {
            get { return y; }
            set
            {
                if (y < 0)
                    throw new ArgumentOutOfRangeException("Location.Y");
                y = value;
            }
        }

        public Location(double x, double y)
        {
            this.X = x;
            this.Y = y;
        }

        public double DistanceTo(Location target)
        {
            return Math.Sqrt(Math.Pow((target.X - this.X), 2) + Math.Pow((target.Y - this.Y), 2));
        }

        public static double Distance(Location start, Location end)
        {
            return start.DistanceTo(end);
        }

        public Location Move(double dx, double dy)
        {
            return new Location(x + dx, y + dy);
        }

        public override string ToString()
        {
            return string.Format("[{0}; {1}]", X, Y);
        }

        public override bool Equals(object obj)
        {
            if (!(obj is Location))
                return false;
            Location second = (Location)obj;
            return (this.X == second.X) && (this.Y == second.Y);
        }
    }

    public class Line
    {
        public Location Start { get; private set; }
        public Location End { get; private set; }

        public double Length { get; private set; }

        public Line(Location start, Location end)
        {
            this.Start = start;
            this.End = end;
            this.Length = start.DistanceTo(end);
        }

        public Location IntersectWith(Line second)
        {
            double param = (Start.X - End.X) * (second.Start.Y - second.End.Y) - (Start.Y - End.Y) * (second.Start.X - second.End.X);
            if (param == 0) //rovnobezne
                return null; // specialni pripad, kdy se budou prekryvat bude resen jinak
            double iX = ((Start.X * End.Y - Start.Y * End.X) * (second.Start.X - second.End.X) - (Start.X - End.X) * (second.Start.X * second.End.Y - second.Start.Y * second.End.X)) / param;
            double iY = ((Start.X * End.Y - Start.Y * End.X) * (second.Start.Y - second.End.Y) - (Start.Y - End.Y) * (second.Start.X * second.End.Y - second.Start.Y * second.End.X)) / param;
            if (((iX >= Start.X && iX <= End.X) || (iX >= End.X && iX <= Start.X)) //ToDo: nahradit nejakou metodou ... Line.Contains(Location) ?
                && ((iY >= Start.Y && iY <= End.Y) || (iY >= End.Y && iY <= Start.Y))
                && ((iX >= second.Start.X && iX <= second.End.X) || (iX >= second.End.X && iX <= second.Start.X))
                && ((iY >= second.Start.Y && iY <= second.End.Y) || (iY >= second.End.Y && iY <= second.Start.Y)))  // pokud prusecik usecek neni az za koncem nektere usecky
                return new Location(iX, iY);
            return null;
        }
    }
}
