﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;

namespace caRoads
{
    class Region
    {
        private Location leftTopCorner;
        private double width;
        private double height;

        public double Coefficient { get; private set; }

        private Rect rectangle;
        private Line[] borderLines;

        public Region(Location corner, double width, double height, double coefficient)
        {
            this.leftTopCorner = corner;
            this.width = width;
            this.height = height;
            this.Coefficient = coefficient;
            rectangle = new Rect(leftTopCorner.X, leftTopCorner.Y, width, height);
            borderLines = new Line[] {
            new Line(leftTopCorner, leftTopCorner.Move(width, 0)),
            new Line(leftTopCorner.Move(width, 0), leftTopCorner.Move(width, -height)),
            new Line(leftTopCorner.Move(width, -height), leftTopCorner.Move(0, -height)),
            new Line(leftTopCorner.Move(0, -height), leftTopCorner)
        };
        }

        public static Region Parse(string text)
        {
            string[] values = text.Split(Globals.separator, StringSplitOptions.RemoveEmptyEntries);
            double x, y, width, height, coefficient;
            if (values.Length != 5
                || !double.TryParse(values[0].Trim(), out coefficient)
                || !double.TryParse(values[1].Trim(), out x)
                || !double.TryParse(values[2].Trim(), out y)
                || !double.TryParse(values[3].Trim(), out width)
                || !double.TryParse(values[3].Trim(), out height)
                )
                throw new FormatException("Vstupni parametr nema spravny format.");
            return new Region(new Location(x, y), width, height, coefficient);
        }

        public bool ContainsLocation(Location point)
        {
            if (point.X >= leftTopCorner.X && point.X <= (leftTopCorner.X + width)
                && point.Y <= leftTopCorner.Y && point.Y >= leftTopCorner.Y - height)
                return true;
            return false;
        }

        public List<Location> GetIntersectionWithRoadPart(RoadPart roadPart)
        {
            List<Location> result = new List<Location>();
            if (this.ContainsLocation(roadPart.Start))
                result.Add(roadPart.Start);
            if (this.ContainsLocation(roadPart.End))
                result.Add(roadPart.End);
            foreach (Line borderLine in borderLines)
            {
                Location intersect = roadPart.IntersectWith(borderLine);
                if ((intersect != null) && (!result.Contains(intersect)))
                    result.Add(intersect);
            }
            return result;
        }


        public IEnumerable<Location> GetIntersections(Road road)
        {
            //ToDo:
            throw new NotImplementedException();
        }
    }
}
