﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace caRoads
{
    public class RoadPart : Line
    {
        //private Location start;
        //private Location end;
        public double Coefficient { get; private set; }
        public double Duration { get; private set; }

        public RoadPart(Location start, Location end, double regionCoefficient = 1)
            : base(start, end)
        {
            //this.start = start;
            //this.end = end;
            //this.regionCoefficient = regionCoefficient;
            this.Coefficient = regionCoefficient;
            this.Duration = this.Length * regionCoefficient;
        }

    }

    public class Road
    {
        public Town TownFrom { get; private set; }
        public Town TownTo { get; private set; }

        public RoadType Type { get; set; }

        private List<RoadPart> roadParts;
        public IEnumerable<RoadPart> RoadParts
        {
            get
            {
                return roadParts;
            }
        }

        public double Length { get; private set; }
        public double Price { get { return Length * Type.Price; } }
        public double Duration
        {
            get
            {
                return this.Type.Speed * roadParts.Sum(r => r.Duration);
            }
        }

        public Road(Town townFrom, Town townTo, RoadType type)
        {
            this.TownFrom = townFrom;
            this.TownTo = townTo;
            this.Type = type;
            this.Length = TownFrom.Location.DistanceTo(TownTo.Location);
            this.roadParts = new List<RoadPart>() { new RoadPart(townFrom.Location, townTo.Location) };
        }

        internal void SplitRoadPart(RoadPart currentRoadPart, List<Location> intersections, double coefficient)
        {
            this.roadParts.Remove(currentRoadPart);
            List<Location> newPoints = new List<Location>();
            newPoints.AddRange(intersections);
            if (!intersections.Contains(currentRoadPart.Start))
                newPoints.Add(currentRoadPart.Start);
            if (!intersections.Contains(currentRoadPart.End))
                newPoints.Add(currentRoadPart.End);
            newPoints = newPoints.OrderBy(i => currentRoadPart.Start.DistanceTo(i)).ToList();
            for (int i = 0; i < newPoints.Count - 1; i++)
            {
                if (intersections.Contains(newPoints[i]) && intersections.Contains(newPoints[i + 1]))
                    this.roadParts.Add(new RoadPart(newPoints[i], newPoints[i + 1], coefficient));
                else
                    this.roadParts.Add(new RoadPart(newPoints[i], newPoints[i + 1], currentRoadPart.Coefficient));
            }
        }

        public override string ToString()
        {
            return string.Format("From: {0} To: {1} Duration: {2:F2} Price: {3:F2}", TownFrom, TownTo, Duration, Price);
        }
    }
}
