﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace caRoads
{
    public class Way
    {
        LinkedList<Town> towns;
        List<Road> roads;
        IEnumerable<Road> availableRoads;

        public Town From
        {
            get
            {
                return towns.First.Value;
            }
        }
        public Town To
        {
            get
            {
                return towns.Last.Value;
            }
        }


        public double Duration
        {
            get
            {
                return roads.Sum(r => r.Duration);
            }
        }
        public double Price
        {
            get
            {
                return roads.Sum(r => r.Price);
            }
        }


        public Way(Town from, IEnumerable<Road> availableRoads)
        {
            towns = new LinkedList<Town>();
            towns.AddFirst(from);
            this.availableRoads = availableRoads;
            roads = new List<Road>();
        }

        private IEnumerable<Way> FindWays(Town to)
        {
            Queue<Way> source = new Queue<Way>();
            source.Enqueue(this);
            List<Way> result = new List<Way>();

            while (source.Count > 0)
            {
                Way currentWay = source.Dequeue();
                if (currentWay.To == to)
                {
                    result.Add(currentWay);
                    break;
                }
                IEnumerable<Road> possibleRoads = availableRoads.Where(r => r.TownFrom == currentWay.To || r.TownTo == currentWay.To);
                foreach (Road road in possibleRoads)
                {
                    Way newWay = Clone(currentWay);
                    if (newWay.AddRoad(road))
                        source.Enqueue(newWay);
                }
            }

            return result;
        }

        private Way Clone(Way currentWay)
        {
            Way result = new Way(currentWay.From, currentWay.availableRoads);
            result.towns.Clear();
            foreach (Town town in currentWay.towns)
                result.towns.AddLast(town);
            foreach (Road road in currentWay.roads)
                result.roads.Add(road);
            return result;
        }

        private bool AddRoad(Road road)
        {
            if (road.TownFrom == this.To && !this.towns.Contains(road.TownTo))
                {
                    this.towns.AddLast(road.TownTo);
                this.roads.Add(road);
                return true;
                }
            if (road.TownTo == this.To && !this.towns.Contains(road.TownFrom))
            {
                this.towns.AddLast(road.TownFrom);
                this.roads.Add(road);
                return true;
            }
            return false;
        }

        public static IEnumerable<Way> GetWays(Town from, Town to, IEnumerable<Road> builtRoads)
        {
            Way way = new Way(from, builtRoads);
            return way.FindWays(to);
        }

        public static bool ExistWay(Town from, Town to, IEnumerable<Road> builtRoads)
        {
            return GetWays(from, to, builtRoads).Count() > 0;
        }

    }
}
