#ifndef VECTOR_H
#define VECTOR_H

#include <cmath>

struct vector2f{
public:
	vector2f(){val[0]=0; val[1]=0;}
	vector2f(int x){val[0]=x; val[1]=x;}
	vector2f(float x){val[0]=x; val[1]=x;}
	vector2f(double x){val[0]=float(x); val[1]=float(x);}
	vector2f(int x, int y){val[0]=x; val[1]=y;}
	vector2f(float x, float y){val[0]=x; val[1]=y;}
	vector2f(double x, double y){val[0]=float(x); val[1]=float(y);}

	inline float length2(){return val[0]*val[0]+val[1]*val[1];}
	inline float length(){return sqrtf(length2());}

	inline vector2f operator+(const vector2f& r)const{return vector2f(val[0] + r[0], val[1] + r[1]);}
	inline vector2f operator-(const vector2f& r)const{return vector2f(val[0] - r[0], val[1] - r[1]);}
	inline vector2f operator*(float f){return vector2f(val[0] * f, val[1] * f);}
	inline vector2f operator/(float f){return vector2f(val[0] / f, val[1] / f);}

	inline bool operator==(vector2f& r){return val[0] == r[0] && val[1] == r[1];}
	inline bool operator!=(vector2f& r){return !operator==(r);}

	inline vector2f& operator+=(vector2f& r){
		(*this)[0] += r[0];
		(*this)[1] += r[1];

		return *this;
	}
	inline vector2f& operator-=(vector2f& r){
		val[0] -= r[0];
		val[1] -= r[1];

		return *this;
	}
	inline vector2f& operator*=(float f){
		val[0] *= f;
		val[1] *= f;

		return *this;
	}
	inline vector2f& operator/=(float f){
		val[0] /= f;
		val[1] /= f;

		return *this;
	}

	float& operator [](unsigned int i){return val[i];}
	float operator [](unsigned int i)const{return val[i];}

	float getX(){return val[0];}
	float getY(){return val[1];}
	void setX(float x){val[0]=x;}
	void setY(float y){val[1]=y;}
	void setX(double x){val[0]=float(x);}
	void setY(double y){val[1]=float(y);}
private:
	float val[2];
};

template<typename T>
class Vector: public vector<T>{
public:

	inline bool includes(T t){
		uint32_t i;
		for(i=0;i<this->size();++i)
			if((*this)[i]==t)return 1;
		return 0;
	}
	inline bool includesSame(T t, uint32_t& index){
		uint32_t i;
		for(i=0;i<this->size();++i){
			if(((*this)[i])->compare(t)){
				index=i;
				return 1;
			}
		}
		return 0;
	}
	inline void addNew(T t){
		if(!includes(t))this->push_back(t);
	}
private:
};

#endif
