#include "world.h"

char errMsg[][50] = {
	"Invalid input file!",//0
	"Invalid city ID!",
	"Invalid city coords!",
	"Invalid road ID!",
	"Invalid road info!",
	"Invalid money data!", //5
	"Invalid condition!",
	"Invalid region info!",
	"Two cities with same ID!",
	"Condition contains invalid city ID!",
	"Two roads with same ID!",//10
	"No road type loaded!",
	"Not enougth city info!",
	"Not enougth road info!",
	"Not enougth condition info!",
	"Not enougth region info!",//15
};

int compareRoadTypeByTime(const void * a, const void * b){ //for qsort
  if((*(RoadType**)a)->getTime() <  (*(RoadType**)b)->getTime())return -1;
  if((*(RoadType**)a)->getTime() == (*(RoadType**)b)->getTime())return 0;
  if((*(RoadType**)a)->getTime() >  (*(RoadType**)b)->getTime())return 1;
}
int compareConditionByTime(const void * a, const void * b){ //for qsort
  if((*(Condition**)a)->getTime() <  (*(Condition**)b)->getTime())return -1;
  if((*(Condition**)a)->getTime() == (*(Condition**)b)->getTime())return 0;
  if((*(Condition**)a)->getTime() >  (*(Condition**)b)->getTime())return 1;
}

World::~World(){
	uint32_t i;
	for(i=0;i<Roads.size();++i)delete Roads[i];
	for(i=0;i<cities.size();++i)delete cities[i];
	for(i=0;i<roadTypes.size();++i)delete roadTypes[i];
	for(i=0;i<regions.size();++i)delete regions[i];
	for(i=0;i<conditions.size();++i)delete conditions[i];
}

char *World::setup(char *inFile){
	FILE *inF;
	uint32_t i,j;
	char temp[1000];
	uint32_t actState=0;
	char *startC;
	uint32_t actP;
	double in;
	int32_t inI;
	double mem[5];
	int32_t memI[4];

	//for sure (if "setup" was called earlier)
	for(i=0;i<cities.size();++i)delete cities[i];
	cities.clear();
	for(i=0;i<roadTypes.size();++i)delete roadTypes[i];
	roadTypes.clear();
	for(i=0;i<regions.size();++i)delete regions[i];
	regions.clear();
	for(i=0;i<conditions.size();++i)delete conditions[i];
	conditions.clear();

	money = 0;

	if(!(inF=fopen(inFile, "r")))
		return errMsg[0];

	
	while(fgets(temp, 1000, inF)){//loading data
		if(temp[0]<0x20)actState++;
		if(actState==0){//cities
			if(temp[0]>='0' && temp[0]<='9'){
				startC=strtok(temp, " :;");
				actP=0;
				while(startC){
					if(actP==0){
						if(strToInt(startC,100,inI,10))
							return errMsg[1];
						memI[0]=inI;
					}
					if(actP==1 || actP==2){
						if(strToDouble(startC,100,in))
							return errMsg[2];
						mem[actP-1]=in;
					}

					startC = strtok(0, " :;");
					++actP;
				}
				if(actP<=2)return errMsg[12];
				cities.push_back(new City(vec2(mem[0], mem[1]), memI[0]));
			}
		}else if(actState==1){//road types
			if(temp[0]>='0' && temp[0]<='9'){
				startC=strtok(temp, " :;");
				actP=0;
				while(startC){
					if(actP==0){
						if(strToInt(startC,100,inI,10))
							return errMsg[3];
						memI[0]=inI;
					}
					if(actP==1 || actP==2){
						if(strToDouble(startC,100,in))
							return errMsg[4];
						mem[actP-1]=in;
					}

					startC = strtok(0, " :;");
					++actP;
				}
				if(actP<=2)return errMsg[13];
				roadTypes.push_back(new RoadType(mem[0], mem[1], memI[0]));
			}
		}else if(actState==2){//money
			if(temp[0]>='0' && temp[0]<='9'){
				if(strToDouble(temp, 100, in))
					return errMsg[5];
				money = float(in);
			}
		}else if(actState==3){//conditions
			if(temp[0]>='0' && temp[0]<='9'){
				startC=strtok(temp, " :;");
				actP=0;
				while(startC){
					if(actP==0 || actP==1){
						if(strToInt(startC,100,inI, 10))
							return errMsg[6];
						memI[actP]=inI;
					}
					if(actP==2){
						if(strToDouble(startC,100,in))
							return errMsg[6];
						mem[0]=in;
					}

					startC = strtok(0, " :;");
					++actP;
				}
				if(actP<=2)return errMsg[14];
				conditions.push_back(new Condition(mem[0], memI[0],memI[1]));
			}
		}else if(actState==4){//regions
			if(temp[0]>='0' && temp[0]<='9' || temp[0]=='.' || temp[0]==','){
				startC=strtok(temp, " :;");
				actP=0;
				while(startC){
					if(actP<5){
						if(strToDouble(startC,100,in))
							return errMsg[7];
						mem[actP]=in;
					}

					startC = strtok(0, " :;");
					++actP;
				}
				if(actP<=4)return errMsg[15];
				regions.push_back(new Region(mem[0], vec2(mem[3]<0?mem[1]+mem[3]:mem[1], mem[4]<0?mem[2]+mem[4]:mem[2]),
						vec2(mem[3]>=0?mem[1]+mem[3]:mem[1], mem[4]>=0?mem[2]+mem[4]:mem[2])));
			}
		}else break;
	}

	//sort for quicker algorithm
	qsort(&roadTypes[0], roadTypes.size(), sizeof(RoadType*), compareRoadTypeByTime);
	qsort(&conditions[0], conditions.size(), sizeof(RoadType*), compareConditionByTime);

	for(i=0;i<conditions.size();++i){//finding which index of "cities" are IDs corresponding
		memI[0]=0;//is found 1. city
		memI[1]=0;//is found 2. city
		memI[2]=0;//newCity1
		memI[3]=0;//newCity2
		for(j=0;j<cities.size();++j){
			if(conditions[i]->getCN(0) == cities[j]->getType()){
				if(memI[0]){
					return errMsg[8];
				}else{
					memI[0]=1;
					memI[2]=j;
				}
			}
			if(conditions[i]->getCN(1) == cities[j]->getType()){
				if(memI[1]){
					return errMsg[8];
				}else{
					memI[1]=1;
					memI[3]=j;
				}
			}
		}
		if(!memI[0] || !memI[1]){
			return errMsg[9];
		}
		conditions[i]->setCN(0,memI[2]);
		conditions[i]->setCN(1,memI[3]);
	}
	if(!roadTypes.size())return errMsg[11];

	for(i=0;i<roadTypes.size();++i){
		for(j=i+1;j<roadTypes.size();++j){
			if(roadTypes[i]->getType() == roadTypes[j]->getType()){
				return errMsg[10];
			}
		}
	}

//	printf("money: %g\n", money);
//	for(i=0;i<cities.size();++i){
//		printf("city: X%g Y%g Type:%d\n", cities[i]->getPos().getX(), cities[i]->getPos().getY(), cities[i]->getType());
//	}
//	for(i=0;i<roadTypes.size();++i){
//		printf("roadType: money:%g time:%g type:%d\n", roadTypes[i]->getMoney(), roadTypes[i]->getTime(), roadTypes[i]->getType());
//	}
//	for(i=0;i<conditions.size();++i){
//	printf("condition: time:%g CityN:%d CityN:%d \n", conditions[i]->getTime(), conditions[i]->getCN(0), conditions[i]->getCN(1));
//	}
//	for(i=0;i<regions.size();++i){
//		printf("region: diff:%g LUX:%g LUY:%g RDX:%g RDY:%g \n", regions[i]->getDiff(), regions[i]->getLU().getX(),
//			regions[i]->getLU().getY(), regions[i]->getRD().getX(), regions[i]->getRD().getY());
//	}

	fclose(inF);
	return 0;
}

//finds cities connected to "city" with roads
//with original city with 0 index
void World::findConCities(uint32_t city, Vector<uint32_t>& con){
	uint32_t i;
	size_t lastSize=0;
	bool incl[2];

	con.clear();
	con.push_back(city);

	while(con.size()!=lastSize){
		lastSize = con.size();
		for(i=0;i<roads.size();++i){
			incl[0] = con.includes(roads[i]->getCN(0));
			incl[1] = con.includes(roads[i]->getCN(1));
			if(incl[0]){
				if(!incl[1])
					con.push_back(roads[i]->getCN(1));
			}else if(incl[1])
				con.push_back(roads[i]->getCN(0));
		}
	}
}

uint8_t World::solveCond(uint32_t actCond, float actMoney){
	Vector<uint32_t> doneCities;
	if(actCond<conditions.size())
		return solveRoad(actCond, conditions[actCond]->getCN(0), conditions[actCond]->getCN(1), 0, actMoney, doneCities);

	return completeRoads(actMoney);
}

//checks whether "actCity" & "endCity" are connected and time by that roads from "actCity" to "endCity" fits in conditions
uint8_t World::checkConInTime(uint32_t actCond, uint32_t actCity, uint32_t endCity, float actTime, Vector<uint32_t>& doneCities){
	uint32_t i;
	if(actTime>=conditions[actCond]->getTime())return 0;
	if(actCity==endCity)return 1;

	for(i=0;i<roads.size();++i){
		if(roads[i]->getCN(0)==actCity){
			if(!doneCities.includes(roads[i]->getCN(1))){
				doneCities.push_back(actCity);
				if(checkConInTime(actCond, roads[i]->getCN(1), endCity, actTime+roads[i]->getTime(), doneCities)){
					doneCities.pop_back();
					return 1;
				}
				doneCities.pop_back();
			}
		}else if(roads[i]->getCN(1)==actCity){
			if(!doneCities.includes(roads[i]->getCN(0))){
				doneCities.push_back(actCity);
				if(checkConInTime(actCond, roads[i]->getCN(0), endCity, actTime+roads[i]->getTime(), doneCities)){
					doneCities.pop_back();
					return 1;
				}
				doneCities.pop_back();
			}
		}
	}
	return 0;
}

//tests all roads from "actCity"
//returns when actCity is endCity(before that tries to find solution for next condition with actual set of roads)
//or when it's more expensive than the best solution(or money limit)
//or when has bigger time than condition allows
uint8_t World::solveRoad(uint32_t actCond,uint32_t actCity,uint32_t endCity,float actTime, float actMoney, Vector<uint32_t>& doneCities){
	uint32_t i,j;
	uint32_t index;
	Vector<uint32_t> con;
	Road *nRoad;
	

	if(actMoney<maxRest || actTime>conditions[actCond]->getTime())return 0;
	if(actCity==endCity)return solveCond(actCond+1, actMoney);

	findConCities(endCity,con);

	if(con.includes(actCity)){
		if(checkConInTime(actCond, actCity, endCity, actTime, doneCities)){
			return solveCond(actCond+1, actMoney);
		}
	}
	//firstly tries to connect cities already connected to endCity
	//it's more likely to be best solution
	for(i=0;i<con.size();++i){
		if(con[i]==actCity || doneCities.includes(con[i]))continue;
		for(j=0;j<roadTypes.size();++j){
			nRoad = getRoad(actCity, con[i],j);
			if(!roads.includesSame(nRoad, index)){
				roads.push_back(nRoad);
				doneCities.push_back(actCity);
				if(solveRoad(actCond, con[i], endCity, actTime+nRoad->getTime(), actMoney-nRoad->getMoney(), doneCities))return 1;
				doneCities.pop_back();
				roads.pop_back();
			}else{
				doneCities.push_back(actCity);
				if(solveRoad(actCond, con[i],endCity,actTime+roads[index]->getTime(), actMoney-roads[index]->getMoney(), doneCities))
					return 1;
				doneCities.pop_back();
				break;
			}
		}
	}
	for(i=0;i<cities.size();++i){
		if(i==actCity || con.includes(i) || doneCities.includes(i))continue;
		for(j=0;j<roadTypes.size();++j){
			nRoad = getRoad(actCity, i,j);
			if(!roads.includesSame(nRoad, index)){
				roads.push_back(nRoad);
				doneCities.push_back(actCity);
				if(solveRoad(actCond, i, endCity, actTime+nRoad->getTime(), actMoney-nRoad->getMoney(), doneCities))return 1;
				doneCities.pop_back();
				roads.pop_back();
			}else{
				doneCities.push_back(actCity);
				if(solveRoad(actCond, i,endCity,actTime+roads[index]->getTime(), actMoney-roads[index]->getMoney(), doneCities))
					return 1;
				doneCities.pop_back();
				break;
			}
		}
	}
	return 0;
}

//tries to connect all cities with each other (through 1 or more roads)
//with cheapest roads (cheapest solution)
uint8_t World::completeRoads(float actMoney){

	uint32_t numRoads=roads.size(); //startering count -> to delete
	uint32_t i,j,k;
	uint32_t cheapR=0; //cheapest road
	Vector<Road *>RestRoads;
	Vector<Vector<uint32_t> > connect;
	for(i=0;i<cities.size();++i)connect.push_back(Vector<uint32_t>());

	for(i=0;i<cities.size();++i){
		for(j=i+1;j<cities.size();++j){
			RestRoads.push_back(getRoad(i, j, cheapRT));
		}
		if(!connect[i].size()){
			findConCities(i,connect[i]);
			for(j=1;j<connect[i].size();++j){
				connect[connect[i][j]].push_back(connect[i][j]);
					
				for(k=0;k<connect[i].size();++k)
					if(k!=j)connect[connect[i][j]].push_back(connect[i][k]);
			}
		}
	}


	if(roads.size()){
		for(i=0;i<RestRoads.size();++i){
			if(connect[RestRoads[i]->getCN(0)].includes(RestRoads[i]->getCN(1))){
				RestRoads.erase(RestRoads.begin()+i);//connects already connected cities
				--i;
			}
		}
	}

	while(RestRoads.size()){
		cheapR=0;
		for(i=1;i<RestRoads.size();++i){
			if(RestRoads[cheapR]->getMoney()>RestRoads[i]->getMoney())cheapR=i;
		}
		roads.push_back(RestRoads[cheapR]);
		actMoney-=RestRoads[cheapR]->getMoney();
		
		RestRoads.erase(RestRoads.begin()+cheapR);

		for(i=0;i<cities.size();++i)connect[i].clear();
		for(i=0;i<cities.size();++i){
			if(!connect[i].size()){
				findConCities(i,connect[i]);
				for(j=1;j<connect[i].size();++j){
					connect[connect[i][j]].push_back(connect[i][j]);

					for(k=0;k<connect[i].size();++k)
						if(k!=j)connect[connect[i][j]].push_back(connect[i][k]);
				}
			}
		}


		for(i=0;i<RestRoads.size();++i){
			if(connect[RestRoads[i]->getCN(0)].includes(RestRoads[i]->getCN(1))){
				RestRoads.erase(RestRoads.begin()+i);//connects already connected cities
				--i;
			}
		}
		if(actMoney<maxRest)break;
	}

	if(actMoney>=maxRest){
		maxRest=actMoney;
		if(info)printf("New cheapeast set of roads! => cost %g\n", money-maxRest);
		roadsEnd.clear();
		for(i=0;i<roads.size();++i)roadsEnd.push_back(roads[i]);
	}

	while(roads.size()>numRoads){
		roads.pop_back();
	}
	if(stop_on_result && actMoney==maxRest)return 1;
	return 0;
}

//returns precomputed roads
Road *World::getRoad(uint32_t c1,uint32_t c2,uint32_t type){
	return Roads[type + c2*roadTypes.size() + c1*roadTypes.size()*cities.size()];
}


//returns 0 for no problem
//returns 1 for not exist result
//returns 2 for problem with file writing

//stop_on_result >
// == 0 => finds best money saving result         > slower
// == 1 => finds first condition matching result  > faster
//info > prints into terminal cost of new best set of roads
//debugInf > prints into file cost, time and length of road
uint8_t World::solve(char *outFile, bool stop_on_result, bool info, bool debugInf){
	uint32_t i,j,k;
	this->stop_on_result = stop_on_result;
	this->info = info;

	//for sure (if "solve" was called earlier)
	roadsEnd.clear();
	roads.clear();

	for(i=0;i<Roads.size();++i){
		delete Roads[i];
	}
	Roads.clear();

	for(i=0;i<cities.size();++i){     //precompute Roads
		for(j=0;j<cities.size();++j){
			for(k=0;k<roadTypes.size();++k){
				Roads.push_back( new Road(i,j,k));
				Roads[Roads.size()-1]->computeTM(roadTypes[k]->getMoney(),roadTypes[k]->getTime(),regions,cities);
			}
		}
	}

	cheapRT=0;
	for(i=1;i<roadTypes.size();++i)
		if(roadTypes[cheapRT]->getMoney()>roadTypes[i]->getMoney())cheapRT=i;

	maxRest = 0;

	solveCond(0,money);

	if(roadsEnd.size()!=0){
		if(printToF(outFile, debugInf))return 2;
		return 0;
	}
	return 1;
}

uint8_t World::printToF(char *outFile, bool debugInf){
	uint32_t i;
	FILE *outF;
	char buff[100];

	if(!(outF=fopen(outFile, "w"))){
		fputs("Invalid output file!\nStopping program!\n",stderr);
		return 1;
	}

	for(i=0;i<roadsEnd.size();++i){
		sprintf(buff,"%d %d %d\n", cities[roadsEnd[i]->getCN(0)]->getType(), cities[roadsEnd[i]->getCN(1)]->getType(),
				roadTypes[roadsEnd[i]->getType()]->getType());
		fputs(buff, outF);
//		computes time & money of road
		if(debugInf){
			sprintf(buff,"money: %g, time: %g, length: %g\n", roadsEnd[i]->getMoney(), roadsEnd[i]->getTime(), roadsEnd[i]->getLen());
			fputs(buff, outF);
		}
	}

	fclose(outF);
	return 0;
}
