#ifndef WORLD_H
#define WORLD_H

#include "common.h"
#include "city.h"
#include "road.h"
#include "roadType.h"
#include "region.h"
#include "condition.h"

class World{
public:
	World(){}
	~World();

	char *setup(char *inFile);  //returns 0 for no problem
	                            //returns char * describing the error

	uint8_t solve(char *outFile, bool stop_on_result=0, bool info=0, bool debugInf=0);
	//returns 0 for no problem
	//returns 1 for not exist result
	//returns 2 for problem with file writing

	//stop_on_result >
	// == 0 => finds best money saving result         > slower
	// == 1 => finds first condition matching result  > faster
	//info > prints into terminal cost of new best set of roads
	//debugInf > prints into file cost, time and length of road

private:
	uint8_t printToF(char *outF, bool debugInf);

	//finds cities connected to "city" with roads
	//with original city with 0 index
	void findConCities(uint32_t city, Vector<uint32_t>& connections);

	//checks whether "actCity" & "endCity" are connected and time by that roads from "actCity" to "endCity" fits in conditions
	uint8_t checkConInTime(uint32_t actCond, uint32_t actCity, uint32_t endCity, float actTime, Vector<uint32_t>& doneCities);
	//tests all roads from "actCity"
	//returns when actCity is endCity(before that tries to find solution for next condition with actual set of roads)
	//or when it's more expensive than the best solution(or money limit)
	//or when has bigger time than condition allows
	uint8_t solveCond(uint32_t actCond, float actMoney);
	//tries to connect all cities with each other (through 1 or more roads)
	//with cheapest roads (cheapest solution)
	uint8_t solveRoad(uint32_t actCond,uint32_t actCity, uint32_t endCity, float actTime, float actMoney, Vector<uint32_t>& doneCities);


	uint8_t completeRoads(float actMoney);

	Road *getRoad(uint32_t c1,uint32_t c2,uint32_t type);//returns precomputed roads

	Vector<Road*> roads;//actRoads

	//best solution for now
	float maxRest;
	Vector<Road*> roadsEnd;

	bool stop_on_result;
	bool info;

	//unchanging data
	Vector<City*> cities;
	Vector<RoadType*> roadTypes;
	Vector<Region*> regions;
	Vector<Condition*> conditions;

	Vector<Road*> Roads; //precomputed roads

	uint32_t cheapRT;//cheapest road type
	float money;
};

#endif
